#  To Do:
#     add improved error checking on XML and XSLT parsing
#     improve XML editing interface
#     templatize HTML further - <tr>s are still embedded
#     improve navigation - next and back, I guess
#     make administration pages more secure
#     write build script for users
#     file-based locking for DBM files?  
#
#  Changelog
#  ---------
#
#  Version 0.01 : 5/4/2000
#     - First version, implemented basic functionality
#
#  Version 0.02 : 5/5/2000
#     - Added a lot of documentation (Hi, Phil!)
#     - Cleaned up a few little things
#
#  Version 0.03 : 5/6/2000
#     - Added ability to edit XSLT stylesheet
#     - User can now edit admin password
#     - Began to convert entries to Lifebox::Entry objects, but I
#       think that might be a bit of overkill as stuff sits now, 
#       so I'm leaving that out for now.  
#     - Base DTD now included with distribution
#     - Added basic search functionality
#     - Fixed some usability stuff on startup
#
#  Version 0.04 : 5/7/2000
#     - Added templating to replace printing HTML
#     - Added ability to edit HTML template
#     - Added ability to edit index page
#     - Minor bug fixes
#
#  Version 0.05 : 5/7/2000
#     - Okay, okay, objectified entry stuff
#     - Changed lifebox.cgi and lbadmin.cgi to deal with objects
#     - Option to edit stylesheet disappeared from lbadmin!  It's back.
#     - I think this'll be the last update for a while, barring bug
#       fixes and/or urgent or interesting user requests.  If anyone
#       finds a problem or wants something, now's the time to ask!
#
#  Version 0.06 : 3/18/2001
#     - Updated stuff so that it works with the new XML::XSLT module
#     - I really need to work on this more often.
#
#  Version 0.07 : 6/8/2001
#     - Adding date indexing functionality to make next/last links work
#     - Hoo boy was that code icky.  I'm sensing a loss of dedication.
#     - Goddamnit, something broke attribute parsing in XML::XSLT.  
#     - Need to fix latest on edit bug, too.  Slacker.

package Lifebox;
BEGIN {push @INC, "/home/majcher/perl/lib"}   #  This is where I have some
					      #  modules installed.  If you
					      #  need to, change this.

use strict;
use CGI qw(:standard);
use Fcntl qw(:DEFAULT :flock);
use DB_File;
use XML::XSLT;
use Text::Template;
#use CGI::Carp qw(fatalsToBrowser);

#
#  Define globals
#

my $version = "0.07";

my $lifeboxdb     = "lifeboxdb";           #  Entry database 
                                           #  --duplicated in Lifebox::Entry
my $configdb      = "lifebox.config";      #  Configuration database
my $stylesheet    = "lifebox.xslt";        #  XSLT stylesheet
my $template_file = "lifebox.template";    #  Text::Template template
my $index_file    = "lifebox.index";       #  Index page HTML
my $entry_index   = "lbindexdb";           #  Entry index database

##
##  Begin Lifebox subroutines
##


#
#  Database functions:  these are for reading, checking, and writing
#     data from the database.  Eventually, these will be replaced 
#     with an object-oriented interface.
#

sub safe_dbmopen {

    my ($hashref, $dbfile, $mode) = @_;
    
    dbmopen(%$hashref, $dbfile, $mode) || die "Can't open $dbfile: $!";

}

sub safe_dbmclose {

    my $hashref = shift;

    dbmclose %$hashref;
}
 
sub goodpass {

    #  Usage:  print "Bad password" unless (Lifebox::goodpass($password));
    #
    #    Takes a string as an argument and returns true if the
    #    password given matches the admin password in the config db.

    my $password = shift;

    my %config;
    safe_dbmopen(\%config, $configdb, 0700);
    my $result = ($password eq $config{password});
    safe_dbmclose \%config;

    return $result;
}


sub get_prefs {
    
    #  Usage:  my %prefs = &Lifebox::get_prefs;
    #
    #     Reads the user preferences from the config database, and
    #     returns them as a hash.

    my %config;
    my %prefs;
    safe_dbmopen(\%config, $configdb, 0700);
    my %prefs = %config;
    safe_dbmclose \%config;

    return %prefs;
}

sub set_prefs {

    #  Usage:  &Lifebox::set_prefs(%prefs);
    #
    #     Takes a hash of user preferences and writes the to the 
    #     config database.
    
    my %prefs = @_ or return;

    my %config;
    safe_dbmopen (\%config, $configdb, 0700);
    for my $key (keys %prefs) {
	$config{$key} = $prefs{$key};
    }
    safe_dbmclose \%config;
}


#  HTML functions:  these print out standard HTML pieces.  Eventually,
#     these will be replaced by Text::Template functionality.
#     (They have been.  Yay!)

sub print_template {

    #  Usage:  &Lifebox::print_template($bodytext);
    #
    #  Takes the bodytext you want to include in the template as
    #  a parameter, inserts it into a standard HTML template, and
    #  prints it out.

    my $bodytext = shift || '&nbsp;';
    my $entry = shift || param('page') || &Lifebox::Entry::latest;

    my %config; 
    safe_dbmopen(\%config, $configdb, 0700);
    my %vars = %config;
    safe_dbmclose \%config;

    $vars{bodytext} = $bodytext;
    $vars{next} = &get_next_key($entry);
    $vars{last} = &get_last_key($entry);

    my $template = Text::Template->new(SOURCE => $template_file)
	or die "Couldn't construct template: $Text::Template::ERROR";
    my $result = $template->fill_in(HASH => \%vars);

    if (defined $result) { 
	print "Content-type: text/html\n\n$result";
    }
    else {
	die "Couldn't fill in template: $Text::Template::ERROR";
    }

}
 

#
#  Display functions:  these process and display the entries in various
#     ways.  The interfaces on these functions can probably be improved.
#


sub display_entry {

    #  Usage:  &Lifebox::display_entry($page);
    #
    #     Takes an entry key as an argument, and returns the entry text.
    
    my $page = shift || param('page') || &Lifebox::Entry::latest;
    my $entry = Lifebox::Entry::read($page);

    return ($entry->process, $page);
}

sub display_latest {

    #  Usage:  &Lifebox::display_latest;
    #
    #     Returns the text of the most recently added entry.

    my $latest_page = &Lifebox::Entry::latest;
    return (display_entry($latest_page), $latest_page);

}

sub display_random {

    #  Usage:  &Lifebox::display_random;
    #
    #     Returns the text of a randomly chosen entry from the database.

    my $random_key = &Lifebox::Entry::random;
    return (display_entry($random_key), $random_key);
}

sub display_search {

    #  Usage:  &Lifebox::display_search;
    #
    #     This will display a form allowing the user to search for
    #     a keyword(s) in the entry database.

    my $term = param('search_term');
    my $content;

    $content = "<tr><td colspan=\"2\"><center><h2>Search Lifebox</h2></center></td></tr>" unless (defined $term);

    $content .= <<SEARCH;
    <form action="lifebox.cgi" method="post">
    <input type="hidden" name="action" value="result">
    <tr><td colspan="2">Search for string:
    <input type="text" name="search_term" size="20" value="$term"><br />
    <input type="submit" value="Search"></td></tr>
    </form>
SEARCH

    return $content;
}

sub display_result {
    
    #  Usage:  &Lifebox::display_result;
    #
    #     This will perform a search on a key, and display the results.
    #     Man, lookit all that HTML in my code!  Ew!

    my $search_term = param('search_term');
    my @result_keys = &Lifebox::Entry::search($search_term);
    my $result_list = "<tr><td colspan=\"2\"><ul>";

    for my $key (@result_keys) {
	my $entry = &Lifebox::Entry::read($key);
	my $time = $entry->time;
	my $title = $entry->title;

	$result_list .= <<RESULTS
   <li> <a href="lifebox.cgi?action=display&page=$time">$title</a></li>
RESULTS
	
    }

    $result_list .= "</ul></td></tr>";

    my $results = <<RESULT;	    
<tr><td colspan="2"><center><h2>Search Results</h2></center></td></tr>
RESULT

    my $matches = scalar(@result_keys);
    if ($matches) {
	$results .= "<tr><td colspan=\"2\"><b>Your search matched $matches pages:</b></td></tr>$result_list";
    }
    else {
	$results .= "<tr><td colspan=\"2\"><b>Your search produced no matches.  Try again.</b></td></tr>";
    }

    $results .= &display_search;

    return $results;
}

#
#  Subsystem functions:  These allow you to manipulate some of the 
#     underlying stuff for the system - templates, etc.
#


sub get_xslt {

    #  Usage:  my $stylesheet = &Lifebox::get_xslt;
    #
    #     Reads the XSLT stylesheet file and returns it as a string

    open(XSLT, $stylesheet) || die "can't open $stylesheet: $!";
    my $content;
    {
	local $/ = undef;    
	$content = <XSLT>;	 
    }
    close XSLT;

    return $content;
    
}

sub write_xslt {

    #  Usage:  &Lifebox::write_xslt($content);
    #
    #     Takes an XSLT stylesheet as a string, and writes it to the 
    #     file.  We're not locking the file when we write to it,
    #     because only the owner should be editing it.  
    
    my $content = shift || return;

    open (XSLT, ">$stylesheet") || die "can't open $stylesheet: $!";
    print XSLT $content;
    close XSLT;        

}

sub get_template {

    #  Usage:  my $template = &Lifebox::get_template;
    #
    #     Reads the HTML temaplate file and returns it as a string

    open(HTML, $template_file) || die "can't open $template_file: $!";
    my $content;
    {
	local $/ = undef;    
	$content = <HTML>;	 
    }
    close HTML;

    return $content;
    
}

sub write_template {

    #  Usage:  &Lifebox::write_template($content);
    #
    #     Takes an HTML template as a string, and writes it to the 
    #     file.  We're not locking the file when we write to it,
    #     because only the owner should be editing it.  
    
    my $content = shift || return;

    open (HTML, ">$template_file") || die "can't open $template_file: $!";
    print HTML $content;
    close HTML;        

}


sub get_index {

    #  Usage:  my $index = &Lifebox::get_index;
    #
    #     Reads the index file and returns it as a string

    open(HTML, $index_file) || die "can't open $index_file: $!";
    my $content;
    {
	local $/ = undef;    
	$content = <HTML>;	 
    }
    close HTML;

    return $content;
    
}

sub write_index {

    #  Usage:  &Lifebox::write_index($content);
    #
    #     Takes an index page as a string, and writes it to the 
    #     file.  We're not locking the file when we write to it,
    #     because only the owner should be editing it.  
    
    my $content = shift || return;

    open (HTML, ">$index_file") || die "can't open $index_file: $!";
    print HTML $content;
    close HTML;        

}

sub index_entries {

    #  Usage:  &Lifebox::index_entries();
    #
    #     Generates a next/last index database from the current
    #     lifebox entries.  

    my (%BOX, %INDEX);
    safe_dbmopen(\%BOX, $lifeboxdb, 0700);
    my @entries = sort {$a<=>$b} &Lifebox::Entry::search('.');
    safe_dbmclose(\%BOX);

    safe_dbmopen(\%INDEX, $entry_index, 0700);
    for (my $i=0; $i<@entries; $i++) {
	#  This will set the "previous" entry for the first entry
	#  to the latest entry, but I think I can live with that.
	#  If it really bugs someone, it's an easy fix.
	$INDEX{$entries[$i]} = join(":",($entries[$i-1],$entries[$i+1]));
    }
    safe_dbmclose(\%INDEX);

}

#  ick.  sloppy.

sub get_next_key {
    my %INDEX;
    my $key = shift || param('page') || &Lifebox::Entry::latest;
    safe_dbmopen(\%INDEX, $entry_index, 0700);
    my ($last,$next) = split(":",$INDEX{$key});
    safe_dbmclose(\%INDEX);
    return $next;
}

sub get_last_key {
    my %INDEX;
    my $key = shift || param('page') || &Lifebox::Entry::latest;
    safe_dbmopen(\%INDEX, $entry_index, 0700);
    my ($last,$next) = split(":",$INDEX{$key});
    safe_dbmclose(\%INDEX);
    return $last;
}


##
##  Object instantiation and method stuff
##
#

package Lifebox::Entry;


#  Define package globals

my $lifeboxdb = "lifeboxdb";            #  Entry database


use Class::Struct;

#  Struct creates 'new' and accessor methods for Entry object

struct 'Lifebox::Entry' => {
    time   => '$',
    title  => '$',
    xml    => '$',
};

sub read {

    #  Usage:  my $entry = &Lifebox::Entry::read($key);
    #
    #     Takes an entry key as an argument, and returns an entry
    #     object.  Returns undef if it can't read the entry.

    my $key = shift or return;

    my %BOX;
    Lifebox::safe_dbmopen(\%BOX, $lifeboxdb, 0700);
    return unless defined $BOX{$key};
    my ($title, $content) = split("\t", $BOX{$key}, 2);
    Lifebox::safe_dbmclose \%BOX;

    my $entry = Lifebox::Entry->new();
    $entry->time($key);
    $entry->title($title);
    $entry->xml($content);

    return $entry;
}

sub write {

    #  Usage:  $entry->write;
    #
    #     Writes the data and content from an entry object back to 
    #     the lifebox database, and updates the latest entry.

    my $self = shift;

    my $time    = $self->time;
    my $title   = $self->title;
    my $content = $self->xml;
    $content =~ s/\t/    /g;

    my %BOX;
    Lifebox::safe_dbmopen(\%BOX, $lifeboxdb, 0700);
    $BOX{$time} = "$title\t$content";
    $BOX{latest} = $time;
    Lifebox::safe_dbmclose \%BOX;
}

sub process {

    #  Usage:  my $content = $entry->process;
    #
    #     Runs and XSLT transformation on the XML in an entry,
    #     and returns the new text as a string.

    my $self = shift;

    my $entry   = $self->time;
    my $time    = localtime($entry);
    my $title   = $self->title;
    my $content = $self->xml;

    #   read font preferences from the config file

    my %config;
    Lifebox::safe_dbmopen(\%config, $configdb, 0700);
    my $font = $config{'font'};
    Lifebox::safe_dbmclose \%config;

    my $title_bar = <<DISP;
<tr>
    <td colspan="2" align="center">
    <font face="$font"><font size="+1"><b>$title</b></font><br>
    <font size="-1">$time</font>
    </td>
</tr>
DISP

#changes to XML::XSLT as of version 0.32
#    my $parser = XML::XSLT->new ($stylesheet, "FILE", warnings => "Active");
#    $parser->transform_document ($content, "STRING");
#    my $content .= $parser->result_string;

    my $parser = XML::XSLT->new(Source => $stylesheet);
    my $content .= $parser->serve(Source => $content, http_headers => 0);
    return $title_bar.$content;	     # for some reason, XML::XSLT is 
                                     # choking when I try to append the
                                     # result_string to $content.  Hrm.
}

sub latest {
    
    #  Usage:  my $latest_key = Lifebox::Entry::latest;
    #
    #     Returns the key of the most recent lifebox entry.

    my %BOX;
    Lifebox::safe_dbmopen(\%BOX, $lifeboxdb, 0700);
    my $key = $BOX{latest};
    Lifebox::safe_dbmclose \%BOX;

    return $key;

}

sub random {

    #  Usage:  my $random_key = Lifebox::Entry::random;
    #
    #     Returns the key of a random lifebox entry.

    my %BOX;
    my $key;
    Lifebox::safe_dbmopen(\%BOX, $lifeboxdb, 0700);
    my @keys = keys %BOX;
    while ($key !~ /^\d+$/) {
	$key = $keys[rand(@keys)];
    }
    Lifebox::safe_dbmclose \%BOX;
    
    return $key;
}

sub pages {

    #  Usage:  my @keys = &Lifebox::Entry::pages;
    #
    #     This will return a list of all the page keys in the database.

    my %BOX;
    Lifebox::safe_dbmopen(\%BOX, $lifeboxdb, 0700);
    my @pagekeys = grep {/^\d+$/} keys %BOX;
    Lifebox::safe_dbmclose \%BOX;

    return @pagekeys;

}

sub search {

    #  Usage:  my @keys = &Lifebox::Entry::search($search_term);
    #
    #     Takes a search keyword as an argument, and returns a list
    #     of keys for entries that match the search, sorted by key.

    my $search_term = shift || return;
    my @result_keys;

    for my $key (sort &Lifebox::Entry::pages) {
	my $entry = &Lifebox::Entry::read($key);
	next unless (($entry->title =~ /$search_term/is) ||
		     ($entry->xml   =~ /$search_term/is));

	#  We have a match!  (My ass and your face!)

	push @result_keys, $key;
    }

    return @result_keys;
}

package Lifebox::Config;

#  TBD

1;

__END__
